/*
 * Program to compute times of the FFT
 * Coro.
 *  modificated for BSP. euroPVM/MPI 99
 * 13-05-98.
 */



/* ----- Libraries ----- */
#include <bsp.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <stdarg.h>


/* ----- Constants ----- */

#define K	(1024)
#define M	(K * K)
#define MaxSIZE	 (2 * M)
#define PI       3.1415926535
#define MAX_TIMES	20

#define PAR(f1, r1, s1, f2, r2, s2) { \
	int ll_partner; \
	\
	NUMPROCESSORS /= 2; \
	BIT--;                            \
	bsp_push_reg(r1,s1);           \
	bsp_push_reg(r2,s2);           \
	bsp_sync();                \
	ll_partner = NAME ^ (1 << BIT);                \
	if ((NAME & (1 << BIT)) == 0) {        \
		f1;                                 \
    bsp_put( ll_partner, r1, r1, 0, s1);\
	}                                     \
	else {                                \
		f2;                                 \
    bsp_put( ll_partner, r2, r2, 0, s2); \
	}                                     \
	bsp_sync();                         \
	NUMPROCESSORS *= 2;                  \
	BIT++;                               \
	bsp_pop_reg(r1);                     \
	bsp_pop_reg(r2);                     \
}

typedef float doubleType;

typedef struct {
	 doubleType re;
	 doubleType im;
} Complex;

/* ----- prototypes ----- */
void write_array(Complex *a);
void initialize(int Size, Complex *a);
void parDandCFFT(Complex *a, Complex *A,int N);
void DandCFFT(Complex *a, Complex *A, int N);

/* ----- Global variables ----- */
int NAME, NUMPROCESSORS, BIT;
Complex a[MaxSIZE], A[MaxSIZE];   /* Array para la seal y su transformada */

/* ----- Functions ----- */

void PRINTF (char *format, ...)
{
  char aux_string[1024];

  va_list vargs;
  va_start(vargs, format);
  aux_string [vsprintf(aux_string, format, vargs)] = '\0';
  va_end(vargs);
  if (NAME == 0)
    printf (aux_string);
}

float diftime(struct timeval *t1,struct timeval *t2) {
  float dif1,dif2;

  dif1 = t2->tv_sec - t1->tv_sec;
  dif2 = dif1 + (t2->tv_usec - t1->tv_usec) / 1.0e6;
  return dif2;
}

/************************************************************/

void initialize (int SIZE, Complex *a) {
	unsigned i;
	for(i = 0; i < SIZE; i++) {
		a[i].re = 1.0;
		a[i].im = 0.0;
	}
}

/*
 * Write results
 */
void write_array(Complex *a) {
	unsigned i;

	for(i = 0; i < 10; i++)
		PRINTF("a[%2u] = [%10f,%10f]\n", i, a[i].re, a[i].im);
}

/*
 * Fast Fourier Transform
 */
void DandCFFT(Complex *a, Complex *A, int N) {
	Complex *b, *B;         /* vectores para los terminos pares  de a[] */
	Complex *c, *C;         /* vectores para los terminos impares de a[] */
	int i, j, n;
	Complex wi;
	int size;

	if (N == 1) {           /* Fase de resolucion del problema trivial */
		A[0].re = a[0].re;
		A[0].im = a[0].im;
	}
	else {
		n = N / 2;
		size = n * sizeof(Complex);
		b = (Complex *)malloc(4 * size);
		if (b == NULL) {
			bsp_abort("Error inside DancCFFT allocating memory...\n");
		}
		B = b + n;
		c = B + n;
		C = c + n;
		/* Fase de division del problema en dos subproblemas de tamanno N/2 */
		for (i = 0, j = 0; i < n; i++, j = 2 * i) {
			b[i] = a[j];        /* b[] almacena las componentes pares de a[] */
			c[i] = a[j+1];      /* c[] almacena las componentes impares de a[] */
		}
		/* Fase de resolucion de los dos subproblemas */
		DandCFFT(b, B, n);
		DandCFFT(c, C, n);
		/* Fase de combinacion de las soluciones de los subproblemas
			 para obtener la solucion del problema original */
		for (i = 0; i < n; i++) {
			wi.re = cos(2.0L * PI * i / N);
			wi.im = sin(2.0L * PI * i / N);
			/* B[i] + w^i C[i]  para las componentes desde 0 hasta N/2-1 */
			A[i].re = B[i].re + (wi.re * C[i].re - wi.im * C[i].im);
			A[i].im = B[i].im + (wi.re * C[i].im + wi.im * C[i].re);
			/* B[i] - w^i C[i] para las componentes desde N/2 hasta N-1 */
			A[i + n].re = B[i].re - (wi.re * C[i].re - wi.im * C[i].im);
			A[i + n].im = B[i].im - (wi.re * C[i].im + wi.im * C[i].re);
		}
		free(b);
	} /* else */
} /* DandCFFT()... */

/*
 * Parallel Fast Fourier Transform
 */
void parDandCFFT(Complex *a, Complex *A, int N) {
	Complex *b, *B;         /* vectores para los terminos pares  de a[] */
	Complex *c, *C;         /* vectores para los terminos impares de a[] */
	int i, j, n;
	Complex wi;
	int size;
	double	t_div_1, t_div_2,
					t_PAR_1, t_PAR_2,
					t_comb_1, t_comb_2,
					t_seqIn_1, t_seqIn_2;

	if(NUMPROCESSORS > 1) {
		if (N == 1) {           /* Fase de resolucion problema trivial */
			A[0].re = a[0].re;
			A[0].im = a[0].im;
		}
		else {
			n = N / 2;
			size = n * sizeof(Complex);
			b = (Complex *)malloc(4 * size);
			if (b == NULL) {
				bsp_abort("Error allocating memory...\n");
			}
			B = b + n;
			c = B + n;
			C = c + n;

			/* Fase de division del problema en dos subproblemas de tamanno N / 2 */
			for (i = 0, j = 0; i < n; i++, j = 2 * i) {
					b[i] = a[j];   /* b[] almacena las componentes pares de a[] */
					c[i] = a[j+1]; /* c[] almacena las componentes impares de a[] */
			}
			/* Fase de resolucion de los dos subproblemas */
			PAR(parDandCFFT(b, B, n), B, size, parDandCFFT(c, C, n), C, size);

			/* Fase de combinacion de las soluciones de los subproblemas
				 para obtener la solucion del problema original */
			for (i = 0; i < n; i++) {
					wi.re = cos(2.0L * PI * i / N);
					wi.im = sin(2.0L * PI * i / N);
					 /* B[i] + w^i C[i]  para las componentes desde 0 hasta N/2-1 */
					A[i].re = B[i].re + (wi.re * C[i].re - wi.im * C[i].im);
					A[i].im = B[i].im + (wi.re * C[i].im + wi.im * C[i].re);
					/* B[i] - w^i C[i] para las componentes desde N/2 hasta N-1 */
					A[i+n].re = B[i].re - (wi.re * C[i].re - wi.im * C[i].im);
					A[i+n].im = B[i].im - (wi.re * C[i].im + wi.im * C[i].re);
			}
			free(b);
		} /* else N == 1... */
	} /* if NUMPROCESSORS ... */
  else DandCFFT(a, A, N);
} /* parDandCFFT()... */

/*
 * Test output
 */
void test_array(int SIZE, Complex *a)
{
	unsigned i, OK = 1;

	if (a[0].re == SIZE)
	{
		 for(i = 1; i < SIZE; i++)
			 if (a[i].re != 0.0 || a[i].im != 0)
			 {
				 OK = 0;
				 break;
			 }
	}
	else OK = 0;
	if (OK) PRINTF("O.K.\n");
	else PRINTF("ERROR\n");
} /* test_array */


main(int argc, char * argv[])
{
	int SIZE, i, REP;
	int sizeInKb;
	int MAX_REP;
	struct timeval time1[1], time2[1];
	double t_seq[MAX_TIMES], t_par[MAX_TIMES];
	Complex *basura;

	bsp_begin(bsp_nprocs());
	NAME = bsp_pid();
	NUMPROCESSORS = bsp_nprocs();
	BIT = log((double)NUMPROCESSORS) / log(2.0);

  if (argc < 3) 
		bsp_abort("need two arguments: size in Kb and MAX_REP");

	sizeInKb = atoi(argv[1]); /* warning! bsprun filename args  is the second? */
	MAX_REP = atoi(argv[2]); 
	/* warning! bsprun filename args  is the second? */
	/* check with Jesus */
	SIZE = K * sizeInKb; /* K * 2^6 = 2^16 = 65.536 */

	initialize(SIZE, a);
	for(REP = 0; REP < MAX_REP; REP++)
	{
		/* ----- Sequential Code ----- */
		bsp_sync();
		if (NAME == 0) { 
				gettimeofday(time1, NULL); 
				DandCFFT(a, A, SIZE); 
				gettimeofday(time2, NULL);
				t_seq[REP] = diftime(time1, time2);
				test_array(SIZE, A);
		}
		/* ----- Parallel Code ----- */
		bsp_sync();
		gettimeofday(time1, NULL);
		parDandCFFT(a, A, SIZE);
		gettimeofday(time2, NULL);
		t_par[REP] = diftime(time1,time2); 
		test_array(SIZE, A);

	} /* for */

	PRINTF("\n+--------------------------------------------------+*\n");
	PRINTF("# \t\t SIZE \t\t SEQ \t\t PAR \t\t SpeedUp \n");
	for(REP = 0; REP < MAX_REP; REP++) {
		PRINTF("#%2d \t %ld \t %9lf \t %9lf \t %9lf\n", 
		NAME, SIZE, t_seq[REP], t_par[REP], (double)(t_seq[REP]/t_par[REP]));
	} /* for */

	PRINTF("\n+--------------------------------------------------+*\n");
	bsp_sync();
	PRINTF("# \t\t SIZE \t\t PAR \n");
	for(REP = 0; REP < MAX_REP; REP++) {
		printf("#%2d \t %ld \t %9lf\n", NAME, SIZE, t_par[REP]);
	} /* for */
	bsp_sync();
	PRINTF("\n+--------------------------------------------------+*\n");

	bsp_end();
} /* main */

